Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

USE work.pkg_081006.all;

----------------------------------------------------
-- This module computes the average of 4 sample, for
-- 4 channels and performs baseline subtraction.

entity Average_4_Samples_v2 is
port(
		clk			: in std_logic;
		nreset		: in std_logic;
		Run_DAQ		: in std_logic; -- resets the averaging process
		
		ADC_STRB	: in std_logic;
		CH1_D		: in std_logic_vector(15 downto 0);                             
		CH2_D		: in std_logic_vector(15 downto 0);                             
		CH3_D		: in std_logic_vector(15 downto 0);                             
		CH4_D		: in std_logic_vector(15 downto 0);                             

		BASE_STRB	: in std_logic; 
		BASE1		: in std_logic_vector(15 downto 0);                             
		BASE2		: in std_logic_vector(15 downto 0);                             
		BASE3		: in std_logic_vector(15 downto 0);                             
		BASE4		: in std_logic_vector(15 downto 0); 
		Sub_Base	: in std_logic;
		RD_BASE		: in std_logic;
		FIFO_FULL   : in std_logic;
		Use_Const_1 : in std_logic;
		Use_Const_2 : in std_logic;
		
		AVG_OUT_STRB : out std_logic;
		AVG1_OUT	 : out std_logic_vector(15 downto 0);                            
		AVG2_OUT	 : out std_logic_vector(15 downto 0);                            
		AVG3_OUT	 : out std_logic_vector(15 downto 0);                            
		AVG4_OUT	 : out std_logic_vector(15 downto 0);
		
		BASE1_OUT	 : out std_logic_vector(15 downto 0);                            
		BASE2_OUT	 : out std_logic_vector(15 downto 0);                            
		BASE3_OUT	 : out std_logic_vector(15 downto 0);                            
		BASE4_OUT	 : out std_logic_vector(15 downto 0);
		
		ADVANCE_ROM  : out std_logic
                           		
);
end Average_4_Samples_v2;

--########################################################
--				A R C H I T E C T U R E
--########################################################
architecture main of Average_4_Samples_v2 is
-- Synchronized Inputs --------------------------------
	signal rst_avg		: std_logic;
	signal get_adc		: std_logic;
	signal halt_daq		: std_logic;
	
	signal daq_active	: std_logic;

-- Control Signals ------------------------------------
	signal cnt_four		: std_logic_vector(2 downto 0);
	signal inc_cnt		: std_logic;
	signal rst_cnt		: std_logic;
	signal sum_done		: std_logic; --added four values
	
	signal clr_sum		: std_logic; -- clear the accumulation
	signal get_sum		: std_logic; -- latch output of adder
	signal get_avg		: std_logic; -- latch final value
	
	signal wr_avg		: std_logic;

-- Input Buffers --------------------------------------
	signal CH1_IN		: std_logic_vector(15 downto 0);
	signal CH2_IN		: std_logic_vector(15 downto 0);
	signal CH3_IN		: std_logic_vector(15 downto 0);
	signal CH4_IN		: std_logic_vector(15 downto 0);
			 
	signal BASE1_IN		: std_logic_vector(15 downto 0);
	signal BASE2_IN		: std_logic_vector(15 downto 0);
	signal BASE3_IN		: std_logic_vector(15 downto 0);
	signal BASE4_IN		: std_logic_vector(15 downto 0);

	signal CH1_SUM		: std_logic_vector(17 downto 0);
	signal CH2_SUM		: std_logic_vector(17 downto 0);
	signal CH3_SUM		: std_logic_vector(17 downto 0);
	signal CH4_SUM		: std_logic_vector(17 downto 0);
	
	signal CH1_OUT		: std_logic_vector(15 downto 0);
	signal CH2_OUT		: std_logic_vector(15 downto 0);
	signal CH3_OUT		: std_logic_vector(15 downto 0);
	signal CH4_OUT		: std_logic_vector(15 downto 0);
	
-- Altera Math Functions ------------------------------
component add
	PORT
	(
		dataa		: IN STD_LOGIC_VECTOR (17 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (17 DOWNTO 0);
		result		: OUT STD_LOGIC_VECTOR (17 DOWNTO 0)
	);
end component;

	signal sum1			: std_logic_vector(17 downto 0);
	signal sum2			: std_logic_vector(17 downto 0);
	signal sum3			: std_logic_vector(17 downto 0);
	signal sum4			: std_logic_vector(17 downto 0);

component subtract
	PORT
	(
		dataa		: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		cout		: OUT STD_LOGIC ;
		result		: OUT STD_LOGIC_VECTOR (15 DOWNTO 0)
	);
end component;

	signal diff1		: std_logic_vector(15 downto 0);
	signal diff2		: std_logic_vector(15 downto 0);
	signal diff3		: std_logic_vector(15 downto 0);
	signal diff4		: std_logic_vector(15 downto 0);

	signal cout1		: std_logic;
	signal cout2		: std_logic;
	signal cout3		: std_logic;
	signal cout4		: std_logic;


-- SETUP_STATE_MACHINE ---------------------------------------
	type STATE_TYPE is
	(init, wait1, get_data, do_sum, ltch_sum, ltch_out, wr_sum);
	signal CS, NS : STATE_TYPE;

--========================================================
--						B E G I N
--========================================================
begin

SYNC_RESET: led 
	port map(sig_in=>Run_DAQ, clk=>clk,	clrn=>nreset, pls_out=>rst_avg); 

SYNC_ADC: led 
	port map(sig_in=>ADC_STRB, clk=>clk,	clrn=>nreset, pls_out=>get_adc); 

SYNC_FIFO_FULL: led 
	port map(sig_in=>FIFO_FULL, clk=>clk,	clrn=>nreset, pls_out=>halt_daq); 

----------------------------------------------------------
INPUT_BUFFERS: 
process(clk, nreset, ADC_STRB)
begin
	if(nreset='0')			then
		CH1_IN		<= (others => '0');	
		CH2_IN		<= (others => '0');			
		CH3_IN		<= (others => '0');			
		CH4_IN		<= (others => '0');			

	elsif(rising_edge(clk))	then
		if(ADC_STRB = '1') then
			CH1_IN		<= CH1_D;	
			CH2_IN		<= CH2_D;			
			CH3_IN		<= CH3_D;			
			CH4_IN		<= CH4_D;
		end if;			   
		 		
	end if;
end process;

BASE_BUFFERS: 
process(clk, nreset, BASE_STRB, Sub_Base)
begin
	if(nreset='0' OR Sub_Base='0') then
		BASE1_IN	<= X"0000";		
		BASE2_IN	<= X"0000";		
		BASE3_IN	<= X"0000";		
		BASE4_IN	<= X"0000";		

	elsif(rising_edge(clk))	then   -- We do not load the baseline once DAQ is running.
		if(BASE_STRB = '1' AND Run_DAQ = '0') then
			BASE1_IN	<= BASE1;		
			BASE2_IN	<= BASE2;		
			BASE3_IN	<= BASE3;		
			BASE4_IN	<= BASE4;		
		end if;           
		
	end if;
end process;
----------------------------------------------------------
CH1SUM: 
add port map(	dataa => "00" & CH1_IN, 
				datab => CH1_SUM, 
				result => sum1);
CH2SUM: 
add port map(	dataa => "00" & CH2_IN, 
				datab => CH2_SUM, 
				result => sum2);
CH3SUM: 
add port map(	dataa => "00" & CH3_IN, 
				datab => CH3_SUM, 
				result => sum3);
CH4SUM: 
add port map(	dataa => "00" & CH4_IN, 
				datab => CH4_SUM, 
				result => sum4);

CH1DIFF: 
subtract port map(	dataa  => CH1_SUM(17 downto 2), 
					datab  => BASE1_IN, 
					cout   => cout1, 
					result => diff1 );
CH2DIFF: 
subtract port map(	dataa  => CH2_SUM(17 downto 2), 
					datab  => BASE2_IN, 
					cout   => cout2, 
					result => diff2 );
CH3DIFF: 
subtract port map(	dataa  => CH3_SUM(17 downto 2), 
					datab  => BASE3_IN, 
					cout   => cout3, 
					result => diff3 );
CH4DIFF: 
subtract port map(	dataa  => CH4_SUM(17 downto 2), 
					datab  => BASE4_IN,
					cout   => cout4, 
					result => diff4 );
					                                                                                         	
----------------------------------------------------------
OUTPUT_BUFFERS: 
process(clk, nreset, clr_sum, get_sum, get_avg)
begin
	if(nreset='0')			then
		CH1_SUM		<= (others => '0');	
		CH2_SUM		<= (others => '0');			
		CH3_SUM		<= (others => '0');			
		CH4_SUM		<= (others => '0');			
		 
		CH1_OUT		<= (others => '0');	
		CH2_OUT		<= (others => '0');			
		CH3_OUT		<= (others => '0');			
		CH4_OUT		<= (others => '0');			
		 
	elsif(rising_edge(clk))	then
		if   (clr_sum = '1') then
			CH1_SUM		<= (others => '0');	
			CH2_SUM		<= (others => '0');			
			CH3_SUM		<= (others => '0');			
			CH4_SUM		<= (others => '0');			
		elsif(get_sum = '1') then
			CH1_SUM		<= sum1;	
			CH2_SUM		<= sum2;			
			CH3_SUM		<= sum3;			
			CH4_SUM		<= sum4;
		end if;			   
		 
		if(get_avg = '1') then
		-- cout is High if for A - B that A > B
			if(cout1 = '1') then 	CH1_OUT	<= diff1;
			else					CH1_OUT <= (others => '0');
			end if;		
			if(cout2 = '1') then 	CH2_OUT	<= diff2;
			else					CH2_OUT <= (others => '0');
			end if;		
			if(cout3 = '1') then 	CH3_OUT	<= diff3;
			else					CH3_OUT <= (others => '0');
			end if;		
			if(cout4 = '1') then 	CH4_OUT	<= diff4;
			else					CH4_OUT <= (others => '0');
			end if;		
		end if;           
		
	end if;
end process;

-----------------------------------------------------------
COUNT_TO_FOUR:
process(clk, nreset, rst_cnt, inc_cnt)
begin
	if  (nreset = '0')			then cnt_four <= (others =>'0');
	elsif(rising_edge(clk))		then
		if   (rst_cnt = '1') 	then cnt_four <= (others =>'0');
		elsif(inc_cnt = '1') 	then cnt_four <= cnt_four + 1;
		else						 cnt_four <= cnt_four;
		end if;			   
	end if;
end process;


--{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{
--                S T A T E   M A C H I N E
--}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}
SM_REG: process(clk, nreset, rst_avg)
begin
	if(nreset='0' OR rst_avg = '1')	then CS <= init;
	elsif(rising_edge(clk))			then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process(CS, rst_avg, get_adc, cnt_four)
begin
	-- NOTE: This process assumes that the Baseline registers have been written
	--       before starting the averaging process.  The averaging process starts
	--		 with Run_DAQ (rst_avg).
	
	-- defaults ----------------------
	clr_sum		<= '0';
	get_sum		<= '0';
	rst_cnt		<= '0';
	inc_cnt		<= '0';
	get_avg		<= '0';
	wr_avg		<= '0';
	
	case CS is
		when init =>					NS <= wait1;	
			clr_sum		<= '1';
			rst_cnt		<= '1';

		when wait1 =>			
			if(get_adc = '1')	then 	NS <= get_data;
			else                   	    NS <= wait1;
			end if;
		
		when get_data =>				NS <= do_sum;
		
		-----------------------------------------------
		when do_sum   =>				NS <= ltch_sum;	
			inc_cnt		<= '1';
		
		when ltch_sum =>				
			get_sum		<= '1';
			
			if(cnt_four < "100")then 	NS <= wait1;
			else                   	    NS <= ltch_out;
			end if;
			
		-----------------------------------------------
		when ltch_out =>				NS <= wr_sum;	
			get_avg		<= '1';
			
		when wr_sum   =>				NS <= init;	
			wr_avg		<= '1';		

		-----------------------------------------------
		when others => 					NS <= init;
	end case;
end process;


STATE_TO_GATE_FIFO_WRITES:	
process(clk, nreset, rst_avg, halt_daq)
begin
	if   (nreset = '0') then daq_active <= '0';
	elsif(rising_edge(clk)) then
		if   (rst_avg  = '1') then daq_active <= '1';
		elsif(halt_daq = '1') then daq_active <= '0';
		end if;
	end if;
end process;

-- OUTPUTS --------------------------------------------------------
	AVG1_OUT <= CH1_OUT;
	AVG2_OUT <= CH2_OUT;
	AVG3_OUT <= CH3_OUT;
	AVG4_OUT <= CH4_OUT;
               process(clk, nreset, RD_BASE)
begin
	if(rising_edge(clk))	then 
		AVG_OUT_STRB <= wr_avg and daq_active;
		ADVANCE_ROM  <= (Use_Const_1 and get_sum) or (Use_Const_2 and wr_avg); -- advance external test vector memory
		
		if(RD_BASE = '0') then  -- latch outputs during a VME read
			BASE1_OUT <= BASE1_IN;
			BASE2_OUT <= BASE2_IN;
			BASE3_OUT <= BASE3_IN;
			BASE4_OUT <= BASE4_IN;
		end if;
		
	end if;
end process;	


end;             

