Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 

------------------------------------------------------
-- MODULE DESCRIPTION
-- 
-- This module does the bulk of the VME address decoding
-- and also implements the status and controls registers
-- for the Booster BLM application.
------------------------------------------------------

entity Integ_Memory is 
	port( 
			clk				: in std_logic;
			nreset			: in std_logic;
			VME_Data_In 	: in std_logic_vector(15 downto 0);
			VME_Addr		: in std_logic_vector(19 downto 0);
			VME_RD_n		: in std_logic;
			VME_WR_n		: in std_logic;
			
			LocalBus_In		: in std_logic_vector(15 downto 0);
			
			FIFO_Data_1		: in std_logic_vector(15 downto 0);
			FIFO_Data_2		: in std_logic_vector(15 downto 0);
			FIFO_Data_3		: in std_logic_vector(15 downto 0);
			FIFO_Data_4		: in std_logic_vector(15 downto 0);
			FIFO_WR_Req		: in std_logic;
			FIFO_Clear		: in std_logic;
			
			Board_ID_Mem	: in std_logic_vector(15 downto 0);
			Serial_Number	: in std_logic_vector(15 downto 0);
			Firmware_Version: in std_logic_vector(15 downto 0);
			Average_Data_In : in std_logic_vector(15 downto 0);
			Base_Data_In	: in std_logic_vector(15 downto 0);
						
			VME_Data_Out	: out std_logic_vector(15 downto 0);
			VME_RD_Sum		: out std_logic; -- control FPGA tri-state output
			
			vmeLD			: out std_logic_vector(15 downto 0);
			vmeLA			: out std_logic_vector(18 downto 0);
			vmeMEM_WE_n		: out std_logic;
			vmeMEM_OE_n		: out std_logic;
			vmeMEM_CS_n		: out std_logic;
			vmeEna_Localbus : out std_logic;
			
			cmd_Start_DAQ 	: out std_logic;
			cmd_Clear_FIFO	: out std_logic;
			Command_Reg_Out : out std_logic_vector(15 downto 0);
			Record_Leng_Out : out std_logic_vector(15 downto 0);
			Alt_Integ_1_Out : out std_logic_vector(15 downto 0);
			Alt_Integ_2_Out : out std_logic_vector(15 downto 0);
			Avg_Length_Out	: out std_logic_vector(15 downto 0);
			Test_DAC_Out	: out std_logic_vector(15 downto 0);
			
			RD_Avg_Data 	: out std_logic_vector(8 downto 1);
			RD_Base_Data 	: out std_logic_vector(4 downto 1);
			Wait_States		: out std_logic_vector(2 downto 0);
			FIFOS_Full		: out std_logic
		);
	end Integ_Memory ;
	
architecture main of Integ_Memory  is
-- VME Read and Write Signals --------------------------------------------
signal VME_RD		: std_logic;
signal VME_WR		: std_logic;

-- Read Pulse ------------------------------------------------------------
signal read_led				: std_logic_vector(1 downto 0);
signal rd_pulse				: std_logic;

-- Application Registers -------------------------------------------------
signal app_register_out		: std_logic_vector(15 downto 0); -- mux of others
signal wr_app_register		: std_logic;

signal command_register		: std_logic_vector(15 downto 0);
signal record_length		: std_logic_vector(15 downto 0);
signal alt_integ_control_1	: std_logic_vector(15 downto 0);
signal alt_integ_control_2	: std_logic_vector(15 downto 0);
signal data_average_length	: std_logic_vector(15 downto 0);
signal test_dac_output		: std_logic_vector(15 downto 0);

signal wr_command_register		: std_logic;
signal wr_record_length			: std_logic;
signal wr_alt_integ_control_1	: std_logic;
signal wr_alt_integ_control_2	: std_logic;
signal wr_data_average_length	: std_logic;
signal wr_test_dac_output		: std_logic;

-- Integrator Data FIFO Memory -------------------------------------------
component BLM_FIFO
	PORT
	(
		clock		: IN STD_LOGIC ;
		data		: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		rdreq		: IN STD_LOGIC ;
		sclr		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		empty		: OUT STD_LOGIC ;
		full		: OUT STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (15 DOWNTO 0);
		usedw		: OUT STD_LOGIC_VECTOR (8 DOWNTO 0)
	);
end component;

signal fifo_status			: std_logic_vector(7 downto 0);

signal wr_pulse_led			: std_logic_vector(1 downto 0);
signal fifo_wr_pulse		: std_logic;

signal rd_fifo_1			: std_logic;
signal rd_fifo_2			: std_logic;
signal rd_fifo_3			: std_logic;
signal rd_fifo_4			: std_logic;

signal rd_fifo_1a			: std_logic;
signal rd_fifo_2a			: std_logic;
signal rd_fifo_3a			: std_logic;
signal rd_fifo_4a			: std_logic;

signal rd_fifo_1b			: std_logic;
signal rd_fifo_2b			: std_logic;
signal rd_fifo_3b			: std_logic;
signal rd_fifo_4b			: std_logic;

signal fifo_1_out			: std_logic_vector(15 downto 0);
signal fifo_2_out			: std_logic_vector(15 downto 0);
signal fifo_3_out			: std_logic_vector(15 downto 0);
signal fifo_4_out			: std_logic_vector(15 downto 0);

signal fifo_1_words			: std_logic_vector(8 downto 0);
signal fifo_2_words			: std_logic_vector(8 downto 0);
signal fifo_3_words			: std_logic_vector(8 downto 0);
signal fifo_4_words			: std_logic_vector(8 downto 0);
           
-- Local Data Bus (upper and lower RAM and lower FPGA registers) ---------
signal rd_local_bus			: std_logic;
signal wr_local_bus			: std_logic;
signal upper_ram_cs			: std_logic;
signal ena_vme_to_localbus	: std_logic;
	
	
	
		
--========================================================================
--========================================================================
--                    B E G I N 
--========================================================================
--========================================================================	
begin

			VME_RD <= not VME_RD_n;
			VME_WR <= not VME_WR_n;
           
READ_PULSE: process(clk, nreset, read_led, VME_RD)
begin
	if   (nreset = '0')	then	read_led <= (others => '0');
	elsif rising_edge(clk) then
		read_led(0) <= VME_RD;
		read_led(1) <= read_led(0);
	end if;
	
	rd_pulse <= VME_RD and not read_led(0);

end process;

WRITE_FIFO_PULSE: process(clk, nreset, wr_pulse_led)
begin
	if   (nreset = '0')	then	wr_pulse_led <= (others => '0');
	elsif rising_edge(clk) then
		wr_pulse_led(0) <= FIFO_WR_Req;
		wr_pulse_led(1) <= wr_pulse_led(0);
	end if;
	
	fifo_wr_pulse <= wr_pulse_led(0) and not wr_pulse_led(1);

end process;
	

TOP_LEVEL_ADDRESS_DECODER: 
process(VME_Addr, Board_ID_Mem, Serial_Number, Firmware_Version, 
		app_register_out, VME_WR, LocalBus_In, fifo_1_out, fifo_2_out, 
		fifo_3_out, fifo_4_out, rd_pulse)
begin
	Wait_States(2 downto 0) <= "001";
	wr_app_register			<= '0';
	ena_vme_to_localbus 	<= '0';
	upper_ram_cs 			<= '0';
	rd_fifo_1b				<= '0';		
	rd_fifo_2b				<= '0';		
	rd_fifo_3b				<= '0';		
	rd_fifo_4b				<= '0';	

	
	if   (VME_Addr < X"00200") then -- Board ID ROM memory
		VME_Data_Out		<= Board_ID_Mem;
		
	elsif(VME_Addr < X"00208") then -- Serial No. ROM memory
		VME_Data_Out		<= Serial_Number;
		
	elsif(VME_Addr < X"00220") then -- version ROM memory
		VME_Data_Out		<= Firmware_Version;
		
	elsif(VME_Addr < X"01000") then -- unused portion
		VME_Data_Out		<= X"bad2";
		
	elsif(VME_Addr < X"01200") then -- Upper FPGA Registers
		VME_Data_Out		<= app_register_out;
		wr_app_register		<= VME_WR;
		
	elsif(VME_Addr < X"01600") then -- FIFO 1 alternate map
		VME_Data_Out		<= fifo_1_out;
		rd_fifo_1b 			<= rd_pulse;
		
	elsif(VME_Addr < X"01A00") then -- FIFO 2 alternate map
		VME_Data_Out		<= fifo_2_out;
		rd_fifo_2b 			<= rd_pulse;
		
	elsif(VME_Addr < X"01E00") then -- FIFO 3 alternate map
		VME_Data_Out		<= fifo_3_out;
		rd_fifo_3b 			<= rd_pulse;
		
	elsif(VME_Addr < X"02200") then -- FIFO 4 alternate map
		VME_Data_Out		<= fifo_4_out;
		rd_fifo_4b 			<= rd_pulse;
		
	elsif(VME_Addr < X"10000") then -- Upper FPGA Memory
		VME_Data_Out		<= X"bad8";
		
	elsif(VME_Addr < X"50000") then -- Upper RAM memory
		Wait_States			<= "010";
		ena_vme_to_localbus <= '1';
		upper_ram_cs 		<= '1';
		VME_Data_Out		<= LocalBus_In;
		
	elsif(VME_Addr < X"80000") then -- Lower FPGA memory
		Wait_States			<= "010";
		ena_vme_to_localbus <= '1';
		VME_Data_Out		<= LocalBus_In;

	elsif(VME_Addr < X"C0000") then -- Lower RAM memory
		Wait_States(2 downto 0) <= "010";
		ena_vme_to_localbus <= '1';
		VME_Data_Out		<= LocalBus_In;
	else
		Wait_States(2 downto 0) <= "001";
		VME_Data_Out		<= X"bad3";
			
	end if;
end process;

-------------------------------------------------------------------
WRITE_APPLICATION_REGISTER: process(VME_Addr, wr_app_register)
begin
	wr_command_register		<= '0';
	cmd_Start_DAQ 			<= '0';
	cmd_Clear_FIFO			<= '0';
	wr_record_length		<= '0';
	wr_alt_integ_control_1 	<= '0';
	wr_alt_integ_control_2 	<= '0';
	wr_data_average_length  <= '0';
	wr_test_dac_output		<= '0';
	
	case VME_Addr is
		when X"01000" => wr_command_register	<= wr_app_register;
		when X"01010" => cmd_Start_DAQ 			<= wr_app_register;
		when X"01014" => cmd_Clear_FIFO			<= wr_app_register;
		when X"01016" => wr_record_length		<= wr_app_register;
		when X"01032" => wr_alt_integ_control_1 <= wr_app_register;
		when X"01034" => wr_alt_integ_control_2 <= wr_app_register;
		when X"01036" => wr_data_average_length <= wr_app_register;
		when X"01048" => wr_test_dac_output		<= wr_app_register;			
		when others => null;
	end case;
end process;

-------------------------------------------------------------------
READ_APPLICATION_REGISTER: 
process(VME_Addr, command_register, record_length, fifo_1_out, rd_pulse,
		fifo_2_out, fifo_3_out, fifo_4_out, fifo_1_words, fifo_2_words,
		fifo_3_words, fifo_4_words, fifo_status, alt_integ_control_1,
		alt_integ_control_2, data_average_length, Average_Data_In,
		VME_RD, test_dac_output, Base_Data_In)
begin
	app_register_out	<= X"bad1";
	rd_fifo_1a			<= '0';		
	rd_fifo_2a			<= '0';		
	rd_fifo_3a			<= '0';		
	rd_fifo_4a			<= '0';	
	
	RD_Avg_Data		<= "00000000"; -- array of select bits	
	RD_Base_Data	<= "0000"; -- array of select bits	
	
	case VME_Addr is
		when X"01000" => app_register_out 	<= command_register; 
		when X"01016" => app_register_out 	<= record_length;
		---------------------------------------------------
		when X"01020" => app_register_out	<= fifo_1_out; 
						 rd_fifo_1a			<= rd_pulse;  
		when X"01022" => app_register_out	<= fifo_2_out; 
						 rd_fifo_2a 		<= rd_pulse;  
		when X"01024" => app_register_out	<= fifo_3_out; 
						 rd_fifo_3a			<= rd_pulse;  
		when X"01026" => app_register_out	<= fifo_4_out; 
						 rd_fifo_4a			<= rd_pulse;  
		----------------------------------------------------
		when X"01028" => 
				app_register_out <= "000000" & fifo_status(0) & fifo_1_words; 
		when X"0102A" => 
				app_register_out <= "000000" & fifo_status(2) & fifo_2_words; 
		when X"0102C" => 
				app_register_out <= "000000" & fifo_status(4) & fifo_3_words; 
		when X"0102E" => 
				app_register_out <= "000000" & fifo_status(6) & fifo_4_words; 
		----------------------------------------------------
		when X"01030" => app_register_out	<= "00000000" & fifo_status;
		when X"01032" => app_register_out 	<= alt_integ_control_1;
		when X"01034" => app_register_out 	<= alt_integ_control_2;
		when X"01036" => app_register_out	<= data_average_length;
		----------------------------------------------------
		when X"01038" => app_register_out	<= Average_Data_In;	
						 RD_Avg_Data(1) 	<= VME_RD;
		when X"0103A" => app_register_out	<= Average_Data_In;	
						 RD_Avg_Data(2) 	<= VME_RD;
		when X"0103C" => app_register_out	<= Average_Data_In;	
						 RD_Avg_Data(3) 	<= VME_RD;
		when X"0103E" => app_register_out	<= Average_Data_In;	
						 RD_Avg_Data(4) 	<= VME_RD;
		when X"01040" => app_register_out	<= Average_Data_In;	
						 RD_Avg_Data(5) 	<= VME_RD;
		when X"01042" => app_register_out	<= Average_Data_In;	
						 RD_Avg_Data(6) 	<= VME_RD;
		when X"01044" => app_register_out	<= Average_Data_In;	
						 RD_Avg_Data(7) 	<= VME_RD;
		when X"01046" => app_register_out	<= Average_Data_In;	
						 RD_Avg_Data(8) 	<= VME_RD;
		-----------------------------------------------------
		when X"01048" => app_register_out 	<= test_dac_output;
		-----------------------------------------------------
		when X"0104A" => app_register_out	<= Base_Data_In;	
						 RD_Base_Data(1)	<= VME_RD;
		when X"0104C" => app_register_out	<= Base_Data_In;	
						 RD_Base_Data(2)	<= VME_RD;
		when X"0104E" => app_register_out	<= Base_Data_In;	
						 RD_Base_Data(3)	<= VME_RD;
		when X"01050" => app_register_out	<= Base_Data_In;	
						 RD_Base_Data(4)	<= VME_RD;				
		when others => null;
	end case;
end process;

--========================================================================
-- 				A P P L I C A T I O N    R E G I S T E R S
--------------------------------------------------------------------------
COMMAND_REG: dffr 
	generic map(16)
	port map(	d    => VME_Data_In,
				clk  => clk, 
				clrn => nreset, 
				ena  => wr_command_register, 
				prn  => '1',
				q    => command_register(15 downto 0)
			);
			 
RECORD_LENGTH_REG: dffr 
	generic map(16)
	port map(	d    => VME_Data_In,
				clk  => clk, 
				clrn => nreset, 
				ena  => wr_record_length, 
				prn  => '1',
				q    => record_length(15 downto 0)
			);

ALT_INTEG_CONTROL_1_REG: dffr 
	generic map(16)
	port map(	d    => VME_Data_In,
				clk  => clk, 
				clrn => nreset, 
				ena  => wr_alt_integ_control_1, 
				prn  => '1',
				q    => alt_integ_control_1(15 downto 0)
			);
			 
ALT_INTEG_CONTROL_2_REG: dffr 
	generic map(16)
	port map(	d    => VME_Data_In,
				clk  => clk, 
				clrn => nreset, 
				ena  => wr_alt_integ_control_2, 
				prn  => '1',
				q    => alt_integ_control_2(15 downto 0)
			);
			 
--DATA_AVG_LENGTH_REG: dffr 
--	generic map(16)
--	port map(	d    => VME_Data_In,
--				clk  => clk, 
--				clrn => nreset, 
--				ena  => wr_data_average_length, 
--				prn  => '1',
--				q    => data_average_length(15 downto 0)
--			);
DATA_AVG_LENGTH_REG: 
process(clk, nreset, wr_data_average_length, VME_Data_In)
begin
	if   (nreset = '0')    then 
			data_average_length <= X"0010";
	elsif rising_edge(clk) then
		if(wr_data_average_length = '1') then
			data_average_length <= VME_Data_In;
		end if;
	end if;
end process;
			 
TEST_DAC_OUTPUT_REG: dffr 
	generic map(16)
	port map(	d    => VME_Data_In,
				clk  => clk, 
				clrn => nreset, 
				ena  => wr_test_dac_output, 
				prn  => '1',
				q    => test_dac_output(15 downto 0)
			);
			 
-- Register Outputs -----------------------------------------------------

	Command_Reg_Out <= command_register;
	Record_Leng_Out <= record_length;
	Alt_Integ_1_Out <= alt_integ_control_1;
	Alt_Integ_2_Out <= alt_integ_control_2;
	Avg_Length_Out	<= data_average_length;
	Test_DAC_Out	<= test_dac_output;

--===========================================================================
--				I N T E G R A T O R    D A T A   F I F O s
-----------------------------------------------------------------------------
-- The fifo's have both a single address assigned in the Upper FPGA register
-- space and an address range assigned in the global address decoding.  The
-- read fifo pulse is driven from both places.  Hence the following summations. 
	rd_fifo_1 <= rd_fifo_1a OR rd_fifo_1b;
	rd_fifo_2 <= rd_fifo_2a OR rd_fifo_2b;
	rd_fifo_3 <= rd_fifo_3a OR rd_fifo_3b;
	rd_fifo_4 <= rd_fifo_4a OR rd_fifo_4b;
                                      
FIFO_CHAN_1:
BLM_FIFO port map(
					clock		=> clk,
					data		=> FIFO_Data_1,
					rdreq		=> rd_fifo_1,
					sclr		=> FIFO_Clear,
					wrreq		=> fifo_wr_pulse,
					empty		=> fifo_status(1),
					full		=> fifo_status(0),
					q			=> fifo_1_out,
					usedw		=> fifo_1_words
				);             

FIFO_CHAN_2:
BLM_FIFO port map(
					clock		=> clk,
					data		=> FIFO_Data_2,
					rdreq		=> rd_fifo_2,
					sclr		=> FIFO_Clear,
					wrreq		=> fifo_wr_pulse,
					empty		=> fifo_status(3),
					full		=> fifo_status(2),
					q			=> fifo_2_out,
					usedw		=> fifo_2_words
				);             

FIFO_CHAN_3:
BLM_FIFO port map(
					clock		=> clk,
					data		=> FIFO_Data_3,
					rdreq		=> rd_fifo_3,
					sclr		=> FIFO_Clear,
					wrreq		=> fifo_wr_pulse,
					empty		=> fifo_status(5),
					full		=> fifo_status(4),
					q			=> fifo_3_out,
					usedw		=> fifo_3_words
				);             

FIFO_CHAN_4:
BLM_FIFO port map(
					clock		=> clk,
					data		=> FIFO_Data_4,
					rdreq		=> rd_fifo_4,
					sclr		=> FIFO_Clear,
					wrreq		=> fifo_wr_pulse,
					empty		=> fifo_status(7),
					full		=> fifo_status(6),
					q			=> fifo_4_out,
					usedw		=> fifo_4_words
				);             
-------------------------------------------------------------------------

-- Local Data Bus Controls (upper and lower RAM, lower FPGA) ------------
	vmeLD			<= VME_Data_In; 
	vmeLA			<= VME_Addr(19 downto 1);	
	vmeMEM_WE_n		<= VME_WR_n;			
	vmeMEM_OE_n		<= VME_RD_n;	
	vmeMEM_CS_n		<= not upper_ram_cs;
	vmeEna_Localbus <= ena_vme_to_localbus;
	
	VME_RD_Sum		<= VME_RD;
    FIFOS_Full		<=  fifo_status(6) OR 
						fifo_status(4) OR
						fifo_status(2) OR
						fifo_status(0);   

end;               