Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 

------------------------------------------------------
-- MODULE DESCRIPTION
-- 

--  TODO:
--  ADD A VME PROCESS TIMEOUT THAT WILL RESET THIS INTERFACE.

-- May 28, 2010 Version 5
--		In version 5 we changed the Interface Ready input
--		logic to a Transfer Done type of logic.

-- January 31, 2011 Version 6
--		In version 6 we brought this VME interface for the
--		Booster LLRF DDS VME module to use in the Booster
--		BLM modules.  A major difference is that the BLM
--		VME interface is only A24-D16 and not A32-D32.
--		So, for this version, the code has been reduced.
--
--		The BLM VME interface changes in that the address
--		decoding for memory on the module is no longer
--		done in the VME interface module.  A separate vhdl
--		module is written for this.
--
-- 		Wait_State input was added to let the address decoding
--		determine the number of wait states needed for the
--		type of memory being addressed.
------------------------------------------------------

entity VME_Interface_v6 is 
	port( 
			clk				: in std_logic;
			nreset			: in std_logic;
			VME_Data_In 	: in std_logic_vector(15 downto 0);
			vmeAddr			: in std_logic_vector(23 downto 1);
			Board_Addr		: in std_logic_vector( 3 downto 0);
			BAM0			: in std_logic;
			BAM1			: in std_logic;
			BAM3			: in std_logic;
			BAM4			: in std_logic;
			BAM5			: in std_logic;
			IACK_n			: in std_logic;
			LWORD_n			: in std_logic;
			AS_n			: in std_logic;
			DS0_n			: in std_logic;
			DS1_n			: in std_logic;
			WRITE_n			: in std_logic;
						
			HoldOff_VME		: in std_logic;
			Wait_States		: in std_logic_vector(2 downto 0);
			
			VME_Data_To_Mem	: out std_logic_vector(15 downto 0);-- version 6
			VME_Addr_To_Mem : out std_logic_vector(19 downto 0);-- version 6
			VME_RD_n	    : out std_logic;
			VME_WR_n	    : out std_logic;
			
			AMATCH_n		: out std_logic;
			CEAB_n 			: out std_logic;
			OEAB_n 			: out std_logic;
			OEBA_n 			: out std_logic;
			
			DTACK			: out std_logic;
			BERR			: out std_logic;

			BS_Out			: out std_logic
			
		);
	end VME_Interface_v6 ;
	
architecture main of VME_Interface_v6  is
-- Address Modifier ------------------------------------------------------
signal  AM				: std_logic_vector(5 downto 0);

-- Syncronized input variables -------------------------------------------
signal  AS				: std_logic;
signal	DS				: std_logic;
signal	WR				: std_logic;
signal  IACK			: std_logic;

signal  sync_BS_d		: std_logic_vector(1 downto 0);
signal  sync_BS_q		: std_logic_vector(1 downto 0);
signal  BS			: std_logic;

-- Variables for combinatorial logic for board addressing ----------------
signal  Addr_Match		: std_logic;
signal	Block_Transfer	: std_logic;
signal  Block_xfer	 	: std_logic; -- syncronized version
signal	Program_Access  : std_logic;
signal  Data_Access		: std_logic;
signal  Standard_Address    : std_logic;

-- Data Transfer Control Signals -----------------------------------------
signal WR_mem			: std_logic;
signal RD_mem			: std_logic;

signal vme_rd_reg		: std_logic; -- these latch the current mode
signal vme_wr_reg		: std_logic;
signal ena_vme_reg		: std_logic;

signal VME_Active_n		: std_logic;
signal Write_nRead		: std_logic;
       
-- Address Latch ---------------------------------------------------------
signal CLR_ADDR_n			: std_logic;
signal LTCH_ADDR			: std_logic;
signal VME_Latched_Addr		: std_logic_vector(19 downto 0);

-- Data Setup Delay Timer (Wait State Timer) -----------------------------
signal timer			: std_logic_vector(2 downto 0);
signal rst_timer		: std_logic;

-- SETUP_STATE_MACHINE ---------------------------------------------------
type STATE_TYPE is
	(idle, check_address, latch_address, data_setup_dly, dtacken, xfer_done);		  
signal CS, NS : STATE_TYPE;
	
--========================================================================
--========================================================================
--                    B E G I N 
--========================================================================
--========================================================================	
begin
--========================================================================			
-- Board Select Logic 

-- STANDARD ADDRESS MODIFIER DECODE ---------------------------------------
-- HEX Value of AM(5..0)	|  Function
--      3F					|  Standard supervisory block transfer
--		3E					|  Standard supervisory program access
--		3D					|  Standard supervisory data access
--		3C					|  Reserved
--      3B					|  Standard nonprivileged block transfer
--		3A					|  Standard nonprivileged program access
--		39					|  Standard nonprivileged data access
--		38					|  Reserved
--
-- "Standard" implies that we are using address bits vmeAddr(23..1).
--
--      0F					|  Extended supervisory block transfer
--		0E					|  Extended supervisory program access
--		0D					|  Extended supervisory data access
--		0C					|  Reserved
--      0B					|  Extended nonprivileged block transfer
--		0A					|  Extended nonprivileged program access
--		09					|  Extended nonprivileged data access
--		08					|  Reserved
--
-- "Extended" implies that we are using address bits vmeAddr(31..1).
-- THE EXTENDED ADDRESS MODE IS NOT SUPPORTED IN THIS VERSION

AM(5 downto 0) 	<= BAM5 & BAM4 & BAM3 & '1' & BAM1 & BAM0;

Block_Transfer	<='1' when AM = X"3F" OR AM = X"3B" else '0';
Program_Access  <='1' when AM = X"3E" OR AM = X"3A" else '0';
Data_Access		<='1' when AM = X"3D" OR AM = X"39" else '0';

Standard_Address <= Block_Transfer OR Program_Access OR Data_Access;

-------------------------------------------------------------------------------
BOARD_ADDRESS_MATCH: 
	process(vmeAddr, Board_Addr, HoldOff_VME, Standard_Address)
	begin	
		if (Standard_Address = '1') then
			if (vmeAddr(23 downto 20) = Board_Addr(3 downto 0)) then 
				Addr_Match <='1'; 
			else 													 
				Addr_Match <='0';
			end if;
		else
			Addr_Match <= '0';
		end if;
	end process;

	 
SYNCRONIZE_VME_INPUTS:	process(clk, nreset, DS0_n, DS1_n, WRITE_n)
	begin
		if(nreset = '0')  then 
			AS		<= '0';
			DS 		<= '0';
			WR 		<= '1';
			IACK	<= '0';
			BS		<= '0';
		elsif rising_edge(clk) then
			AS		<=  NOT AS_n;
			DS 		<= (NOT DS0_n) AND (NOT DS1_n); -- allow only even byte access.
			WR 		<=  NOT WRITE_n;
			IACK	<=  NOT IACK_n;	
			BS		<= ((NOT IACK) AND (NOT AS_n) AND Addr_Match); 						
		end if;
	end process;

SYNCRONIZE_BLOCK_XFER: process(clk, nreset, BS, Block_Transfer)	
	begin
		if(nreset = '0')  then 
			Block_xfer <= '0';
		elsif rising_edge(clk) then
			Block_xfer <= BS AND Block_Transfer;
		end if;
	end process;		

DATA_SETUP_DELAY_TIMER: process(clk, nreset, rst_timer)
	begin
		if   (nreset = '0' OR rst_timer = '1') 	then timer <= (others => '0');
		elsif rising_edge(clk) 					then timer <= timer + 1;
		end if;
	end process;

--==========================================================================
-- VME Address Decoding and Data Source/Destination Selection
-- Something to Realize !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
-- The address memory map given in the documentation for the crate processor 
-- programs includes an A[0] bit which is always assumed to be 0.                                   

ADDRESS_LATCH: dffr 
	generic map(20)
	port map(	d    => vmeAddr(19 downto 1) & '0',
				clk  => clk, 
				clrn => CLR_ADDR_n, 
				ena  => LTCH_ADDR, 
				prn  => '1',
				q    => VME_Latched_Addr(19 downto 0)
			); 

--===========================================================================
--===========================================================================
-- STATE MACHINE  
-------------------------------------------------------------------------
SM_REG: 
process(clk, nreset)
begin
	if(nreset='0')			then CS <= idle;
	elsif(rising_edge(clk))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process(CS, AS, BS, DS, WR, HoldOff_VME, Block_xfer, timer, Wait_States)	
begin
	-- DEFAULT OUTPUT VALUES --------------------------
	CLR_ADDR_n	<= '1';
	LTCH_ADDR 	<= '0';
	DTACK 		<= '0';
	BERR  		<= '0';
	
    RD_mem		<= '0';
	WR_mem		<= '0';
	--rst_vme_reg	<= '0';
	ena_vme_reg	<= '0';
	
	rst_timer 	<= '1';
	-------------------------------------------------------------------------
    
	case CS is			
	-------------------------------------------------------------------------
	--                    INITIAL STATE
		when idle =>		
			CLR_ADDR_n	<= '0';
			
			if (HoldOff_VME  = '1') then 	NS <= idle;
			elsif (AS = '1') 		then 	NS <= check_address;
			else                   			NS <= idle;
			end if;
	-------------------------------------------------------------------------
	--              VME ADDRESS LATCHED IN UPPER FPGA
			
		when check_address => 
			-- This state allows examination of the Board Select (AMATCH) bits.
			LTCH_ADDR     	<= '1'; -- this is a DFF enable.
			
			if    (AS = '0')      	then 	NS <= idle; --just a glitch
			elsif (BS = '1')      	then	NS <= latch_address;
			else  							NS <= idle; 
			end if;
				
		when latch_address => 				
			-- Address bits vmeA[19..1] are latched.
			ena_vme_reg		<= '1';
			
			if (WR = '1') 	then	WR_mem <= '1';
			else					RD_mem <= '1';						
			end if;
							
			if    (BS = '0')      		then 	NS <= idle; --just a glitch
			elsif (HoldOff_VME  = '1')  then  	NS <= latch_address;
			elsif (DS = '1')      		then	NS <= data_setup_dly; 
			else  								NS <= latch_address;
			end if;
			
		when data_setup_dly => 					
			-- Need this delay for memory data to setup.
			rst_timer 		<= '0';
			
			if (WR = '1') 	then	WR_mem <= '1';
			else					RD_mem <= '1';						
			end if;
							
			if   (timer = Wait_States) 	then	NS <= dtacken;
			else								NS <= data_setup_dly;
			end if;
			
		when dtacken => 
			DTACK <= '1'; -- signal host that module is ready
			
			if (DS = '0')         	then  	NS <= xfer_done;
			else							NS <= dtacken; --wait
			end if;
				
		when xfer_done => 			
			
			if    (BS = '0')	  		then 	NS <= idle; 
										-- address strobe goes away
			elsif (HoldOff_VME = '1') 	then 	NS <= xfer_done;
			elsif (DS = '0')	  		then 	NS <= xfer_done;
			elsif (Block_xfer = '1') 	then 	NS <= data_setup_dly; 
				--go if DS='1' (goes active) and block transfer while 
				--board is still selected and VME is not held off (BUSY).
			else					   			NS <= xfer_done;
			end if;

		when others =>              		NS <= idle;
	end case;
end process;


-- Outputs -----------------------------------------------------------------			
VME_Data_To_Mem	<= VME_Data_In;
VME_Addr_To_Mem <= VME_Latched_Addr;

VME_RD_n 	<= not vme_rd_reg;
CEAB_n 		<= not vme_rd_reg;
OEAB_n 		<= not vme_rd_reg;

VME_WR_n	<= not (WR_mem and vme_wr_reg);
OEBA_n 		<= not vme_wr_reg;

--Extend the RD_mem pulse and the write mode OEBA_n signal an additional clock
EXTEND_PULSES: process(clk, nreset, RD_mem, WR_mem, DS0_n, DS1_n, ena_vme_reg)
begin
	if (nreset = '0' OR DS0_n = '1' OR DS1_n = '1')then
		vme_rd_reg <= '0';
		vme_wr_reg <= '0';
	elsif rising_edge(clk)	then
		if (ena_vme_reg = '1') then
			vme_rd_reg  <= RD_mem;
			vme_wr_reg	<= WR_mem;
		end if;		
	end if;
end process;
--DTACK			<= SEE STATE MACHINE;
--BERR			<= SEE STATE MACHINE;
			
AMATCH_n		<= not Addr_Match;
BS_Out 			<= BS; -- diagnostic output

	
end;